from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, overload

if TYPE_CHECKING:
    from typing_extensions import TypeVar

    from kontainer.container import Maybe, Result
    from kontainer.core.types import Container

    ValueT = TypeVar("ValueT", infer_variance=True)
    ValueT0 = TypeVar("ValueT0", infer_variance=True)
    ValueT1 = TypeVar("ValueT1", infer_variance=True)
    ValueT2 = TypeVar("ValueT2", infer_variance=True)
    ValueT3 = TypeVar("ValueT3", infer_variance=True)
    ValueT4 = TypeVar("ValueT4", infer_variance=True)
    ValueT5 = TypeVar("ValueT5", infer_variance=True)
    ValueT6 = TypeVar("ValueT6", infer_variance=True)
    ValueT7 = TypeVar("ValueT7", infer_variance=True)
    ValueT8 = TypeVar("ValueT8", infer_variance=True)
    ValueT9 = TypeVar("ValueT9", infer_variance=True)
    ValueT10 = TypeVar("ValueT10", infer_variance=True)

    @overload
    def compose_funcs() -> Callable[[ValueT], ValueT]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0], /
    ) -> Callable[[ValueT], ValueT0]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0], func1: Callable[[ValueT0], ValueT1], /
    ) -> Callable[[ValueT], ValueT1]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        /,
    ) -> Callable[[ValueT], ValueT2]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        /,
    ) -> Callable[[ValueT], ValueT3]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        /,
    ) -> Callable[[ValueT], ValueT4]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        func5: Callable[[ValueT4], ValueT5],
        /,
    ) -> Callable[[ValueT], ValueT5]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        func5: Callable[[ValueT4], ValueT5],
        func6: Callable[[ValueT5], ValueT6],
        /,
    ) -> Callable[[ValueT], ValueT6]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        func5: Callable[[ValueT4], ValueT5],
        func6: Callable[[ValueT5], ValueT6],
        func7: Callable[[ValueT6], ValueT7],
        /,
    ) -> Callable[[ValueT], ValueT7]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        func5: Callable[[ValueT4], ValueT5],
        func6: Callable[[ValueT5], ValueT6],
        func7: Callable[[ValueT6], ValueT7],
        func8: Callable[[ValueT7], ValueT8],
        /,
    ) -> Callable[[ValueT], ValueT8]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        func5: Callable[[ValueT4], ValueT5],
        func6: Callable[[ValueT5], ValueT6],
        func7: Callable[[ValueT6], ValueT7],
        func8: Callable[[ValueT7], ValueT8],
        func9: Callable[[ValueT8], ValueT9],
        /,
    ) -> Callable[[ValueT], ValueT9]: ...

    @overload
    def compose_funcs(
        func0: Callable[[ValueT], ValueT0],
        func1: Callable[[ValueT0], ValueT1],
        func2: Callable[[ValueT1], ValueT2],
        func3: Callable[[ValueT2], ValueT3],
        func4: Callable[[ValueT3], ValueT4],
        func5: Callable[[ValueT4], ValueT5],
        func6: Callable[[ValueT5], ValueT6],
        func7: Callable[[ValueT6], ValueT7],
        func8: Callable[[ValueT7], ValueT8],
        func9: Callable[[ValueT8], ValueT9],
        func10: Callable[[ValueT9], ValueT10],
        /,
    ) -> Callable[[ValueT], ValueT10]: ...

    @overload
    def compose_funcs(*funcs: Callable[[Any], Any]) -> Callable[[Any], Any]: ...

    @overload
    def compose_bind_funcs() -> Callable[[ValueT], Container[ValueT, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]], /
    ) -> Callable[[ValueT], Maybe[ValueT0]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT1]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT2]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT3]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT4]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        func5: Callable[[ValueT4], Maybe[ValueT5]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT5]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        func5: Callable[[ValueT4], Maybe[ValueT5]],
        func6: Callable[[ValueT5], Maybe[ValueT6]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT6]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        func5: Callable[[ValueT4], Maybe[ValueT5]],
        func6: Callable[[ValueT5], Maybe[ValueT6]],
        func7: Callable[[ValueT6], Maybe[ValueT7]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT7]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        func5: Callable[[ValueT4], Maybe[ValueT5]],
        func6: Callable[[ValueT5], Maybe[ValueT6]],
        func7: Callable[[ValueT6], Maybe[ValueT7]],
        func8: Callable[[ValueT7], Maybe[ValueT8]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT8]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        func5: Callable[[ValueT4], Maybe[ValueT5]],
        func6: Callable[[ValueT5], Maybe[ValueT6]],
        func7: Callable[[ValueT6], Maybe[ValueT7]],
        func8: Callable[[ValueT7], Maybe[ValueT8]],
        func9: Callable[[ValueT8], Maybe[ValueT9]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT9]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Maybe[ValueT0]],
        func1: Callable[[ValueT0], Maybe[ValueT1]],
        func2: Callable[[ValueT1], Maybe[ValueT2]],
        func3: Callable[[ValueT2], Maybe[ValueT3]],
        func4: Callable[[ValueT3], Maybe[ValueT4]],
        func5: Callable[[ValueT4], Maybe[ValueT5]],
        func6: Callable[[ValueT5], Maybe[ValueT6]],
        func7: Callable[[ValueT6], Maybe[ValueT7]],
        func8: Callable[[ValueT7], Maybe[ValueT8]],
        func9: Callable[[ValueT8], Maybe[ValueT9]],
        func10: Callable[[ValueT9], Maybe[ValueT10]],
        /,
    ) -> Callable[[ValueT], Maybe[ValueT10]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]], /
    ) -> Callable[[ValueT], Result[ValueT0, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT1, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT2, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT3, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT4, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        func5: Callable[[ValueT4], Result[ValueT5, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT5, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        func5: Callable[[ValueT4], Result[ValueT5, Any]],
        func6: Callable[[ValueT5], Result[ValueT6, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT6, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        func5: Callable[[ValueT4], Result[ValueT5, Any]],
        func6: Callable[[ValueT5], Result[ValueT6, Any]],
        func7: Callable[[ValueT6], Result[ValueT7, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT7, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        func5: Callable[[ValueT4], Result[ValueT5, Any]],
        func6: Callable[[ValueT5], Result[ValueT6, Any]],
        func7: Callable[[ValueT6], Result[ValueT7, Any]],
        func8: Callable[[ValueT7], Result[ValueT8, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT8, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        func5: Callable[[ValueT4], Result[ValueT5, Any]],
        func6: Callable[[ValueT5], Result[ValueT6, Any]],
        func7: Callable[[ValueT6], Result[ValueT7, Any]],
        func8: Callable[[ValueT7], Result[ValueT8, Any]],
        func9: Callable[[ValueT8], Result[ValueT9, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT9, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Result[ValueT0, Any]],
        func1: Callable[[ValueT0], Result[ValueT1, Any]],
        func2: Callable[[ValueT1], Result[ValueT2, Any]],
        func3: Callable[[ValueT2], Result[ValueT3, Any]],
        func4: Callable[[ValueT3], Result[ValueT4, Any]],
        func5: Callable[[ValueT4], Result[ValueT5, Any]],
        func6: Callable[[ValueT5], Result[ValueT6, Any]],
        func7: Callable[[ValueT6], Result[ValueT7, Any]],
        func8: Callable[[ValueT7], Result[ValueT8, Any]],
        func9: Callable[[ValueT8], Result[ValueT9, Any]],
        func10: Callable[[ValueT9], Result[ValueT10, Any]],
        /,
    ) -> Callable[[ValueT], Result[ValueT10, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]], /
    ) -> Callable[[ValueT], Container[ValueT0, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT1, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT2, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT3, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT4, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        func5: Callable[[ValueT4], Container[ValueT5, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT5, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        func5: Callable[[ValueT4], Container[ValueT5, Any]],
        func6: Callable[[ValueT5], Container[ValueT6, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT6, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        func5: Callable[[ValueT4], Container[ValueT5, Any]],
        func6: Callable[[ValueT5], Container[ValueT6, Any]],
        func7: Callable[[ValueT6], Container[ValueT7, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT7, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        func5: Callable[[ValueT4], Container[ValueT5, Any]],
        func6: Callable[[ValueT5], Container[ValueT6, Any]],
        func7: Callable[[ValueT6], Container[ValueT7, Any]],
        func8: Callable[[ValueT7], Container[ValueT8, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT8, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        func5: Callable[[ValueT4], Container[ValueT5, Any]],
        func6: Callable[[ValueT5], Container[ValueT6, Any]],
        func7: Callable[[ValueT6], Container[ValueT7, Any]],
        func8: Callable[[ValueT7], Container[ValueT8, Any]],
        func9: Callable[[ValueT8], Container[ValueT9, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT9, Any]]: ...

    @overload
    def compose_bind_funcs(
        func0: Callable[[ValueT], Container[ValueT0, Any]],
        func1: Callable[[ValueT0], Container[ValueT1, Any]],
        func2: Callable[[ValueT1], Container[ValueT2, Any]],
        func3: Callable[[ValueT2], Container[ValueT3, Any]],
        func4: Callable[[ValueT3], Container[ValueT4, Any]],
        func5: Callable[[ValueT4], Container[ValueT5, Any]],
        func6: Callable[[ValueT5], Container[ValueT6, Any]],
        func7: Callable[[ValueT6], Container[ValueT7, Any]],
        func8: Callable[[ValueT7], Container[ValueT8, Any]],
        func9: Callable[[ValueT8], Container[ValueT9, Any]],
        func10: Callable[[ValueT9], Container[ValueT10, Any]],
        /,
    ) -> Callable[[ValueT], Container[ValueT10, Any]]: ...

    @overload
    def compose_bind_funcs(
        *funcs: Callable[[Any], Maybe[Any]]
    ) -> Callable[[Any], Maybe[Any]]: ...

    @overload
    def compose_bind_funcs(
        *funcs: Callable[[Any], Result[Any, Any]]
    ) -> Callable[[Any], Result[Any, Any]]: ...

    @overload
    def compose_bind_funcs(
        *funcs: Callable[[Any], Container[Any, Any]]
    ) -> Callable[[Any], Container[Any, Any]]: ...


__all__ = ["compose_funcs", "compose_bind_funcs"]


class _Compose:
    __slots__ = ("_funcs",)

    def __init__(self, funcs: tuple[Callable[[Any], Any], ...]) -> None:
        self._funcs = funcs

    def __call__(self, value: Any) -> Any:
        result = self._funcs[0](value)
        for func in self._funcs[1:]:
            result = func(result)
        return result


class _ComposeBinds:
    __slots__ = ("_funcs",)

    def __init__(self, funcs: tuple[Callable[[Any], Container[Any, Any]], ...]) -> None:
        self._funcs = funcs

    def __call__(self, value: Any) -> Container[Any, Any]:
        result = self._funcs[0](value)
        for func in self._funcs[1:]:
            result = result.bind_value(func)
        return result


def compose_funcs(*funcs: Callable[[Any], Any]) -> Callable[[Any], Any]:
    if not funcs:
        from kontainer.flow.flowtools import identity

        funcs = (identity,)
    return _Compose(funcs)


def compose_bind_funcs(
    *funcs: Callable[[Any], Container[Any, Any]]
) -> Callable[[Any], Container[Any, Any]]:
    if not funcs:
        from kontainer.container.maybe import Some

        funcs = (Some,)
    return _ComposeBinds(funcs)



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Generic, Literal, cast, overload

from typing_extensions import Self, TypeGuard, TypeVar, override

from kontainer.core.const import Undefined, undefined
from kontainer.core.exception import KontainerTypeError
from kontainer.core.types import Container

ValueT = TypeVar("ValueT", infer_variance=True)
OtherT = TypeVar("OtherT", infer_variance=True)
if TYPE_CHECKING:
    ElementT = TypeVar("ElementT", infer_variance=True)

__all__ = ["Maybe"]


class Maybe(Container[ValueT, None], Generic[ValueT]):
    if TYPE_CHECKING:

        @override
        def __copy__(self) -> Maybe[ValueT]: ...

        @override
        def __deepcopy__(self, memo: dict[Any, Any] | None = None) -> Maybe[ValueT]: ...

    @override
    def __init__(self, value: ValueT | None) -> None:
        self._value = value

    @overload
    def __new__(cls, value: Undefined | None) -> Maybe[Any]: ...

    @overload
    def __new__(cls, value: ValueT) -> Maybe[ValueT]: ...

    @overload
    def __new__(cls, value: ValueT | Undefined | None) -> Maybe[ValueT]: ...

    @override
    def __new__(cls, value: ValueT | Undefined | None) -> Maybe[ValueT]:
        if value is None or value is undefined:
            return Null(None)
        value = cast("ValueT", value)
        return Some(value)

    @override
    def __repr__(self) -> str:
        container_type = type(self)
        if issubclass(container_type, Some):
            name = "Some"
        elif issubclass(container_type, Null):
            name = "Null"
        else:
            name = "Maybe"  # pragma: no cover
        return f"<{name}: value={self._value!r}>"

    @override
    def __str__(self) -> str:
        return str(self._value)

    @override
    def __eq__(self, other: object) -> bool:
        return isinstance(other, Maybe) and other._value == self._value

    @override
    def __hash__(self) -> int:
        return hash((Maybe, self._value))

    @override
    def map_value(self, func: Callable[[ValueT], OtherT]) -> Maybe[OtherT]:
        raise NotImplementedError

    @override
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], OtherT]
    ) -> Maybe[OtherT]:
        raise NotImplementedError

    @override
    def map_container(
        self, value: Maybe[ElementT], func: Callable[[ValueT, ElementT], OtherT]
    ) -> Maybe[OtherT]:
        raise NotImplementedError

    @override
    def bind_value(self, func: Callable[[ValueT], Maybe[OtherT]]) -> Maybe[OtherT]:
        raise NotImplementedError

    @override
    def bind_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], Maybe[OtherT]]
    ) -> Maybe[OtherT]:
        raise NotImplementedError

    @override
    def bind_container(
        self, value: Maybe[ElementT], func: Callable[[ValueT, ElementT], Maybe[OtherT]]
    ) -> Maybe[OtherT]:
        raise NotImplementedError

    @override
    def switch(self) -> Maybe[None]:
        raise NotImplementedError

    @override
    def default(self, value: OtherT) -> ValueT | OtherT:
        raise NotImplementedError

    @override
    def map_default(self, func: Callable[[], OtherT]) -> ValueT | OtherT:
        raise NotImplementedError

    @override
    def unwrap(self) -> ValueT:
        raise NotImplementedError

    @property
    @override
    def is_positive(self) -> bool:
        raise NotImplementedError

    @staticmethod
    def some(value: OtherT) -> Some[OtherT]:
        return Some(value)

    @staticmethod
    def null(value: Any) -> Null[Any]:  # noqa: ARG004
        return Null(None)

    @staticmethod
    def is_some(maybe: Maybe[OtherT]) -> TypeGuard[Some[OtherT]]:
        return maybe.is_positive

    @staticmethod
    def is_null(maybe: Maybe[OtherT]) -> TypeGuard[Null[OtherT]]:
        return maybe.is_negative

    @override
    def copy(self) -> Maybe[ValueT]:
        raise NotImplementedError

    @override
    def deepcopy(self) -> Maybe[ValueT]:
        raise NotImplementedError

    @override
    def ensure_positive(self) -> Some[ValueT]:
        raise NotImplementedError

    @override
    def ensure_negative(self) -> Null[ValueT]:
        raise NotImplementedError


class Some(Maybe[ValueT], Generic[ValueT]):
    if TYPE_CHECKING:

        @override
        def __copy__(self) -> Some[ValueT]: ...

        @override
        def __deepcopy__(self, memo: dict[Any, Any] | None = None) -> Some[ValueT]: ...

    _value: ValueT

    @override
    def __new__(cls, value: ValueT) -> Self:
        return super(Container, cls).__new__(cls)

    @override
    def map_value(self, func: Callable[[ValueT], OtherT]) -> Maybe[OtherT]:
        return Some(func(self._value))

    @override
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], OtherT]
    ) -> Maybe[OtherT]:
        return Some(func(self._value, value))

    @override
    def map_container(
        self, value: Maybe[ElementT], func: Callable[[ValueT, ElementT], OtherT]
    ) -> Maybe[OtherT]:
        return value.bind_value(lambda x: self.map_values(x, func))

    @override
    def bind_value(self, func: Callable[[ValueT], Maybe[OtherT]]) -> Maybe[OtherT]:
        return func(self._value)

    @override
    def bind_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], Maybe[OtherT]]
    ) -> Maybe[OtherT]:
        return func(self._value, value)

    @override
    def bind_container(
        self, value: Maybe[ElementT], func: Callable[[ValueT, ElementT], Maybe[OtherT]]
    ) -> Maybe[OtherT]:
        return value.bind_value(lambda x: self.bind_values(x, func))

    @override
    def switch(self) -> Maybe[None]:
        return Null(None)

    @override
    def default(self, value: Any) -> ValueT:
        return self._value

    @override
    def map_default(self, func: Callable[[], Any]) -> ValueT:
        return self._value

    @override
    def unwrap(self) -> ValueT:
        return self._value

    @property
    @override
    def is_positive(self) -> Literal[True]:
        return True

    @override
    def copy(self) -> Some[ValueT]:
        return self.some(self._value)

    @override
    def deepcopy(self) -> Some[ValueT]:
        new = deepcopy(self._value)
        return self.some(new)

    @override
    def ensure_positive(self) -> Some[ValueT]:
        return Some(self._value)

    @override
    def ensure_negative(self) -> Null[ValueT]:
        error_msg = f"{self!r} is not null"
        raise KontainerTypeError(error_msg)


class Null(Maybe[ValueT], Generic[ValueT]):
    if TYPE_CHECKING:

        @override
        def __copy__(self) -> Null[ValueT]: ...

        @override
        def __deepcopy__(self, memo: dict[Any, Any] | None = None) -> Null[ValueT]: ...

    @override
    def __init__(self, value: ValueT | None) -> None:
        if value is undefined:
            value = None
        super().__init__(value)

    @override
    def __new__(cls, value: ValueT) -> Null[Any]:
        return super(Container, cls).__new__(cls)

    @override
    def map_value(self, func: Callable[[ValueT], OtherT]) -> Null[OtherT]:
        return Null(None)

    @override
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], OtherT]
    ) -> Maybe[OtherT]:
        return Null(None)

    @override
    def map_container(
        self, value: Maybe[ElementT], func: Callable[[ValueT, ElementT], OtherT]
    ) -> Maybe[OtherT]:
        return Null(None)

    @override
    def bind_value(self, func: Callable[[ValueT], Maybe[OtherT]]) -> Maybe[OtherT]:
        return Null(None)

    @override
    def bind_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], Maybe[OtherT]]
    ) -> Maybe[OtherT]:
        return Null(None)

    @override
    def bind_container(
        self, value: Maybe[ElementT], func: Callable[[ValueT, ElementT], Maybe[OtherT]]
    ) -> Maybe[OtherT]:
        return Null(None)

    @override
    def switch(self) -> Maybe[None]:
        return Some(None)

    @override
    def default(self, value: OtherT) -> OtherT:
        return value

    @override
    def map_default(self, func: Callable[[], OtherT]) -> OtherT:
        return func()

    @override
    def unwrap(self) -> None:
        return None

    @property
    @override
    def is_positive(self) -> Literal[False]:
        return False

    @override
    def copy(self) -> Null[ValueT]:
        return self.null(self._value)

    @override
    def deepcopy(self) -> Null[ValueT]:
        new = deepcopy(self._value)
        return self.null(new)

    @override
    def ensure_positive(self) -> Some[ValueT]:
        error_msg = f"{self!r} is not some"
        raise KontainerTypeError(error_msg)

    @override
    def ensure_negative(self) -> Null[ValueT]:
        return Null(self._value)



"""https://github.com/dbrattli/Expression/blob/main/expression/collections/seq.py"""

from __future__ import annotations

import builtins
import functools
import itertools
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generator,
    Generic,
    Iterable,
    Iterator,
    overload,
)

from typing_extensions import Concatenate, ParamSpec, TypeVar, Unpack

SourceT = TypeVar("SourceT", infer_variance=True)
if TYPE_CHECKING:
    from kontainer import Maybe

    ResultT = TypeVar("ResultT", infer_variance=True)
    StateT = TypeVar("StateT", infer_variance=True)
    ParamT = ParamSpec("ParamT")
    _T1 = TypeVar("_T1", infer_variance=True)
    _T2 = TypeVar("_T2", infer_variance=True)
    _T3 = TypeVar("_T3", infer_variance=True)
    _T4 = TypeVar("_T4", infer_variance=True)

__all__ = [
    "curry",
    "curry2",
    "curry3",
    "append",
    "delay",
    "fold",
    "head",
    "mapi",
    "scan",
    "skip",
    "tail",
    "take",
    "unfold",
    "zip",
    "identity",
    "first",
    "second",
    "third",
    "fourth",
]


def curry(
    func: Callable[Concatenate[_T1, ParamT], ResultT]
) -> Callable[[_T1], Callable[ParamT, ResultT]]:
    def outer(arg1: _T1, /) -> Callable[ParamT, ResultT]:
        def inner(*args: ParamT.args, **kwargs: ParamT.kwargs) -> ResultT:
            return func(arg1, *args, **kwargs)

        return inner

    return outer


def curry2(
    func: Callable[Concatenate[_T1, _T2, ParamT], ResultT]
) -> Callable[[_T1], Callable[[_T2], Callable[ParamT, ResultT]]]:
    def outer(arg1: _T1, /) -> Callable[[_T2], Callable[ParamT, ResultT]]:
        def outer2(arg2: _T2, /) -> Callable[ParamT, ResultT]:
            def inner(*args: ParamT.args, **kwargs: ParamT.kwargs) -> ResultT:
                return func(arg1, arg2, *args, **kwargs)

            return inner

        return outer2

    return outer


def curry3(
    func: Callable[Concatenate[_T1, _T2, _T3, ParamT], ResultT]
) -> Callable[[_T1], Callable[[_T2], Callable[[_T3], Callable[ParamT, ResultT]]]]:
    def outer(
        arg1: _T1, /
    ) -> Callable[[_T2], Callable[[_T3], Callable[ParamT, ResultT]]]:
        def outer2(arg2: _T2, /) -> Callable[[_T3], Callable[ParamT, ResultT]]:
            def outer3(arg3: _T3, /) -> Callable[ParamT, ResultT]:
                def inner(*args: ParamT.args, **kwargs: ParamT.kwargs) -> ResultT:
                    return func(arg1, arg2, arg3, *args, **kwargs)

                return inner

            return outer3

        return outer2

    return outer


class SeqGen(Iterable[SourceT], Generic[SourceT]):
    """Sequence from a generator function.

    We use this to allow multiple iterations over the same sequence
    generated by a generator function.
    """

    def __init__(self, gen: Callable[[], Iterable[SourceT]]) -> None:
        self.gen = gen

    def __iter__(self) -> Iterator[SourceT]:
        xs = self.gen()
        return builtins.iter(xs)


class Infinite(Iterable[SourceT], Generic[SourceT]):
    """An infinite iterable.

    Where each iterator starts counting at 0.
    """

    def __init__(self, initializer: Callable[[int], SourceT]) -> None:
        self.initializer = initializer

    def __iter__(self) -> Iterator[SourceT]:
        return builtins.map(self.initializer, itertools.count(0, 1))


def append(
    *others: Iterable[SourceT]
) -> Callable[[Iterable[SourceT]], Iterable[SourceT]]:
    """Append sequence to other sequences.

    Wraps the given enumerations as a single concatenated enumeration.
    """
    return functools.partial(_append, others=others)


def delay(generator: Callable[[], Iterable[SourceT]]) -> Iterable[SourceT]:
    """Delay sequence.

    Returns a sequence that is built from the given delayed
    specification of a sequence.

    The input function is evaluated each time an Iterator for the
    sequence is requested.

    Args:
        generator: The generating function for the sequence.
    """
    return SeqGen(generator)


def fold(
    source: Iterable[SourceT],
    folder: Callable[[StateT, SourceT], StateT],
    state: StateT,
) -> StateT:
    """Fold elements in sequence.

    Applies a function to each element of the collection,
    threading an accumulator argument through the computation. If
    the input function is f and the elements are i0...iN then
    computes f (... (f s i0)...) iN.

    Args:
        source: The input sequence to fold.
        folder: A function that updates the state with each element
            from the sequence.
        state: The initial state.

    Returns:
        Partially applied fold function that takes a source sequence and
        returns the state object after the folding function is applied
        to each element of the sequence.
    """
    return functools.reduce(folder, source, state)


def head(source: Iterable[SourceT]) -> SourceT:
    """Return the first element of the sequence.

    Args:
        source: The input sequence.

    Returns:
        The first element of the sequence.

    Raises:
        Raises `ValueError` if the source sequence is empty.
    """
    for value in source:
        return value

    raise ValueError("Sequence contains no elements")


@overload
def init_infinite() -> Iterable[int]: ...


@overload
def init_infinite(initializer: Callable[[int], SourceT]) -> Iterable[SourceT]: ...


@overload
def init_infinite(initializer: None) -> Iterable[int]: ...


@overload
def init_infinite(
    initializer: Callable[[int], SourceT] | None
) -> Iterable[SourceT]: ...


def init_infinite(
    initializer: Callable[[int], SourceT] | None = None
) -> Iterable[SourceT] | Iterable[int]:
    """Generate infinite sequence.

    Generates a new sequence which, when iterated, will return
    successive elements by calling the given function. The results of
    calling the function will not be saved, that is the function will be
    reapplied as necessary to regenerate the elements. The function is
    passed the index of the item being generated.

    Iteration can continue up to `sys.maxsize`.
    """

    if initializer is None:
        return Infinite(_self)
    return Infinite(initializer)


@overload
def starmap(
    mapper: Callable[[_T1, _T2], ResultT]
) -> Callable[[Iterable[tuple[_T1, _T2]]], Iterable[ResultT]]: ...


@overload
def starmap(
    mapper: Callable[[_T1, _T2, _T3], ResultT]
) -> Callable[[Iterable[tuple[_T1, _T2, _T3]]], Iterable[ResultT]]: ...


@overload
def starmap(
    mapper: Callable[[_T1, _T2, _T3, _T4], ResultT]
) -> Callable[[Iterable[tuple[_T1, _T2, _T3, _T4]]], Iterable[ResultT]]: ...


def starmap(
    mapper: Callable[..., ResultT]
) -> Callable[[Iterable[Any]], Iterable[ResultT]]:
    """Starmap source sequence.

    Unpack arguments grouped as tuple elements. Builds a new collection
    whose elements are the results of applying the given function to the
    unpacked arguments to each of the elements of the collection.

    Args:
        mapper: A function to transform items from the input sequence.

    Returns:
        Partially applied map function.
    """

    return functools.partial(_starmap, func=mapper)


def map2(
    mapper: Callable[[_T1, _T2], ResultT]
) -> Callable[[Iterable[tuple[_T1, _T2]]], Iterable[ResultT]]:
    return starmap(mapper)


def map3(
    mapper: Callable[[_T1, _T2, _T3], ResultT]
) -> Callable[[Iterable[tuple[_T1, _T2, _T3]]], Iterable[ResultT]]:
    return starmap(mapper)


def mapi(
    source: Iterable[SourceT], mapping: Callable[[int, SourceT], ResultT]
) -> Iterable[ResultT]:
    """Map list with index.

    Builds a new collection whose elements are the results of
    applying the given function to each of the elements of the
    collection. The integer index passed to the function indicates
    the index (from 0) of element being transformed.

    Args:
        source: The input sequence to to map.
        mapping: The function to transform elements and their
            indices.

    Returns:
        The list of transformed elements.
    """
    return itertools.starmap(mapping, builtins.enumerate(source))


def scan(
    source: Iterable[SourceT],
    scanner: Callable[[StateT, SourceT], StateT],
    state: StateT,
) -> Iterable[StateT]:
    """Scan elements in sequence.

    Like fold, but computes on-demand and returns the sequence of
    intermediary and final results.

    Args:
        source: The input sequence.
        scanner: A function that updates the state with each element
        state: The initial state.
    """
    return itertools.accumulate(source, scanner, initial=state)


def skip(source: Iterable[SourceT], count: int) -> Iterable[SourceT]:
    """Skip elements from sequence.

    Returns a sequence that skips N elements of the underlying
    sequence and then yields the remaining elements of the sequence.

    Args:
        source: The input sequence.
        count: The number of items to skip.
    """

    if count <= 0:
        return source

    return SeqGen(functools.partial(_gen_skip, source, count))


def tail(source: Iterable[SourceT]) -> Iterable[SourceT]:
    """Return tail of sequence.

    Returns a sequence that skips 1 element of the underlying sequence
    and then yields the remaining elements of the sequence.
    """
    return skip(source, 1)


def take(source: Iterable[SourceT], count: int) -> Iterable[SourceT]:
    """Returns the first N elements of the sequence.

    Args:
        source: The source sequence.
        count: The number of items to take.

    Returns:
        The result sequence.
    """

    if count <= 0:
        return ()

    return SeqGen(functools.partial(_gen_take, source, count))


def unfold(
    state: StateT, generator: Callable[[StateT], Maybe[tuple[SourceT, StateT]]]
) -> Iterable[SourceT]:
    """Unfold sequence.

    Generates a list that contains the elements generated by the given
    computation. The given initial state argument is passed to the
    element generator.

    Args:
        state: The initial state.
        generator: A function that takes in the current state and
            returns an option tuple of the next element of the list
            and the next state value.

    Returns:
        A partially applied unfold function that takes the state and
        returns the result list.
    """
    while True:
        result = generator(state)
        value = result.unwrap()
        if value is None:
            break

        item, state = value
        yield item


def zip(  # noqa: A001
    source1: Iterable[SourceT]
) -> Callable[[Iterable[ResultT]], Iterable[tuple[SourceT, ResultT]]]:
    """Zip sequence with other.

    Combines the two sequences into a list of pairs. The two
    sequences need not have equal lengths: when one sequence is
    exhausted any remaining elements in the other sequence are
    ignored.

    Args:
        source1: The first input sequence.

    Returns:
        Partially applied zip function.
    """

    return functools.partial(_zip, source1)


def identity(value: SourceT) -> SourceT:
    return value


def first(values: tuple[SourceT, Unpack[tuple[Any, ...]]]) -> SourceT:
    return values[0]


def second(values: tuple[Any, SourceT, Unpack[tuple[Any, ...]]]) -> SourceT:
    return values[1]


def third(values: tuple[Any, Any, SourceT, Unpack[tuple[Any, ...]]]) -> SourceT:
    return values[2]


def fourth(values: tuple[Any, Any, Any, SourceT, Unpack[tuple[Any, ...]]]) -> SourceT:
    return values[3]


def _starmap(
    args: Iterable[tuple[Any, ...]], func: Callable[[Unpack[tuple[Any, ...]]], Any]
) -> Iterable[Any]:
    for arg in args:
        yield func(*arg)


def _self(x: SourceT) -> SourceT:
    return x


def _append(
    source: Iterable[SourceT], /, others: Iterable[Iterable[SourceT]]
) -> Iterable[SourceT]:
    return _concat(source, *others)


def _concat(*iterables: Iterable[SourceT]) -> Iterable[SourceT]:
    return SeqGen(functools.partial(_gen_concat, iterables))


def _gen_concat(iterables: Iterable[Iterable[SourceT]]) -> Generator[SourceT, Any, Any]:
    for it in iterables:
        yield from it


def _gen_skip(source: Iterable[SourceT], count: int) -> Generator[SourceT, Any, Any]:
    for i, n in enumerate(source):
        if i >= count:
            yield n


def _gen_take(source: Iterable[SourceT], count: int) -> Generator[SourceT, Any, Any]:
    for i, n in enumerate(source):
        yield n

        if i == count - 1:
            break


def _zip(
    source1: Iterable[SourceT], source2: Iterable[ResultT]
) -> Iterable[tuple[SourceT, ResultT]]:
    return builtins.zip(source1, source2)



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Generic, Literal, NoReturn, overload

from typing_extensions import Self, TypeGuard, TypeVar, override

from kontainer.core.const import Undefined, undefined
from kontainer.core.exception import KontainerTypeError
from kontainer.core.types import Container

ValueT = TypeVar("ValueT", infer_variance=True)
OtherT = TypeVar("OtherT", infer_variance=True)
if TYPE_CHECKING:
    ErrorT = TypeVar("ErrorT", infer_variance=True, bound=Exception)
    AnotherT = TypeVar("AnotherT", infer_variance=True)
    ElementT = TypeVar("ElementT", infer_variance=True)

__all__ = ["Result"]


class Result(Container[ValueT, OtherT], Generic[ValueT, OtherT]):
    if TYPE_CHECKING:

        @override
        def __copy__(self) -> Result[ValueT, OtherT]: ...

        @override
        def __deepcopy__(
            self, memo: dict[Any, Any] | None = None
        ) -> Result[ValueT, OtherT]: ...

    @property
    def _value_or_other(self) -> ValueT | OtherT:
        raise NotImplementedError

    @override
    def __init__(self, value: ValueT) -> None:
        self._value = value

    @overload
    def __new__(cls, value: ErrorT) -> Result[Any, ErrorT]: ...

    @overload
    def __new__(cls, value: type[ErrorT]) -> Result[Any, type[ErrorT]]: ...

    @overload
    def __new__(cls, value: ValueT) -> Result[ValueT, Any]: ...

    @overload
    def __new__(
        cls, value: ValueT | ErrorT | type[ErrorT]
    ) -> Result[ValueT, Any] | Result[Any, ErrorT] | Result[Any, type[ErrorT]]: ...

    @override
    def __new__(
        cls, value: ValueT | ErrorT | type[ErrorT]
    ) -> Result[ValueT, Any] | Result[Any, ErrorT] | Result[Any, type[ErrorT]]:
        if isinstance(value, Exception) or (
            isinstance(value, type) and issubclass(value, Exception)
        ):
            return Error(value)
        return Done(value)

    @override
    def __repr__(self) -> str:
        container_type = type(self)
        if issubclass(container_type, Done):
            name = "Done"
        elif issubclass(container_type, Error):
            name = "Error"
        else:
            name = "Result"  # pragma: no cover
        return f"<{name}: value={self._value!r}>"

    @override
    def __str__(self) -> str:
        return str(self._value)

    @override
    def __eq__(self, other: object) -> bool:
        return isinstance(other, Result) and other._value == self._value

    @override
    def __hash__(self) -> int:
        return hash((Result, self._value))

    @override
    def map_value(self, func: Callable[[ValueT], AnotherT]) -> Result[AnotherT, OtherT]:
        raise NotImplementedError

    @override
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Result[AnotherT, OtherT]:
        raise NotImplementedError

    @override
    def map_container(
        self, value: Result[ElementT, Any], func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Result[AnotherT, OtherT]:
        raise NotImplementedError

    @override
    def bind_value(
        self, func: Callable[[ValueT], Result[AnotherT, OtherT]]
    ) -> Result[AnotherT, OtherT]:
        raise NotImplementedError

    @override
    def bind_values(
        self,
        value: ElementT,
        func: Callable[[ValueT, ElementT], Result[AnotherT, OtherT]],
    ) -> Result[AnotherT, OtherT]:
        raise NotImplementedError

    @override
    def bind_container(
        self,
        value: Result[ElementT, Any],
        func: Callable[[ValueT, ElementT], Result[AnotherT, OtherT]],
    ) -> Result[AnotherT, OtherT]:
        raise NotImplementedError

    @override
    def switch(self) -> Result[OtherT, ValueT]:
        raise NotImplementedError

    @override
    def default(self, value: AnotherT) -> ValueT | AnotherT:
        raise NotImplementedError

    @override
    def map_default(self, func: Callable[[], AnotherT]) -> ValueT | AnotherT:
        raise NotImplementedError

    @override
    def unwrap(self) -> ValueT:
        raise NotImplementedError

    def unwrap_error(self) -> NoReturn:
        raise NotImplementedError

    def unwrap_error_or(self, value: AnotherT) -> AnotherT | NoReturn:
        raise NotImplementedError

    def unwrap_error_or_else(self, func: Callable[[], AnotherT]) -> AnotherT | NoReturn:
        raise NotImplementedError

    def unwrap_error_or_with(
        self, func: Callable[[OtherT], AnotherT]
    ) -> AnotherT | NoReturn:
        raise NotImplementedError

    @property
    @override
    def is_positive(self) -> bool:
        raise NotImplementedError

    @staticmethod
    def done(value: AnotherT) -> Done[AnotherT, Any]:
        return Done(value)

    @staticmethod
    def error(value: AnotherT) -> Error[Any, AnotherT]:
        return Error(value)

    @staticmethod
    def is_done(
        result: Result[AnotherT, ElementT]
    ) -> TypeGuard[Done[AnotherT, ElementT]]:
        return result.is_positive

    @staticmethod
    def is_error(
        result: Result[ElementT, AnotherT]
    ) -> TypeGuard[Error[ElementT, AnotherT]]:
        return result.is_negative

    @override
    def copy(self) -> Result[ValueT, OtherT]:
        raise NotImplementedError

    @override
    def deepcopy(self) -> Result[ValueT, OtherT]:
        raise NotImplementedError

    @override
    def ensure_positive(self) -> Done[ValueT, OtherT]:
        raise NotImplementedError

    @override
    def ensure_negative(self) -> Error[ValueT, OtherT]:
        raise NotImplementedError


class Done(Result[ValueT, OtherT], Generic[ValueT, OtherT]):
    if TYPE_CHECKING:

        @override
        def __copy__(self) -> Done[ValueT, OtherT]: ...

        @override
        def __deepcopy__(
            self, memo: dict[Any, Any] | None = None
        ) -> Done[ValueT, OtherT]: ...

    @property
    @override
    def _value_or_other(self) -> ValueT:
        return self._value

    @override
    def __new__(cls, value: ValueT) -> Self:
        return super(Container, cls).__new__(cls)

    @override
    def map_value(self, func: Callable[[ValueT], AnotherT]) -> Result[AnotherT, OtherT]:
        return Done(func(self._value))

    @override
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Result[AnotherT, OtherT]:
        return Done(func(self._value, value))

    @override
    def map_container(
        self, value: Result[ElementT, Any], func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Result[AnotherT, OtherT]:
        return value.bind_value(lambda x: self.map_values(x, func))

    @override
    def bind_value(
        self, func: Callable[[ValueT], Result[AnotherT, OtherT]]
    ) -> Result[AnotherT, OtherT]:
        return func(self._value)

    @override
    def bind_values(
        self,
        value: ElementT,
        func: Callable[[ValueT, ElementT], Result[AnotherT, OtherT]],
    ) -> Result[AnotherT, OtherT]:
        return func(self._value, value)

    @override
    def bind_container(
        self,
        value: Result[ElementT, Any],
        func: Callable[[ValueT, ElementT], Result[AnotherT, OtherT]],
    ) -> Result[AnotherT, OtherT]:
        return value.bind_value(lambda x: self.bind_values(x, func))

    @override
    def switch(self) -> Result[OtherT, ValueT]:
        return Error(self._value)

    @override
    def default(self, value: Any) -> ValueT:
        return self._value

    @override
    def map_default(self, func: Callable[[], Any]) -> ValueT:
        return self._value

    @override
    def unwrap(self) -> ValueT:
        return self._value

    @override
    def unwrap_error(self) -> NoReturn:
        raise KontainerTypeError("Not an error container")

    @override
    def unwrap_error_or(self, value: AnotherT) -> AnotherT:
        return value

    @override
    def unwrap_error_or_else(self, func: Callable[[], AnotherT]) -> AnotherT:
        return func()

    def unwrap_error_or_with(
        self,
        func: Callable[[OtherT], Any],  # noqa: ARG002
    ) -> NoReturn:
        raise KontainerTypeError("Not an error container")

    @property
    @override
    def is_positive(self) -> Literal[True]:
        return True

    @override
    def copy(self) -> Done[ValueT, OtherT]:
        return self.done(self._value)

    @override
    def deepcopy(self) -> Done[ValueT, OtherT]:
        new = deepcopy(self._value)
        return self.done(new)

    @override
    def ensure_positive(self) -> Done[ValueT, OtherT]:
        return Done(self._value)

    @override
    def ensure_negative(self) -> Error[ValueT, OtherT]:
        error_msg = f"{self!r} is not error"
        raise KontainerTypeError(error_msg)


class Error(Result[ValueT, OtherT], Generic[ValueT, OtherT]):
    if TYPE_CHECKING:

        @override
        def __copy__(self) -> Error[ValueT, OtherT]: ...

        @override
        def __deepcopy__(
            self, memo: dict[Any, Any] | None = None
        ) -> Error[ValueT, OtherT]: ...

    __slots__ = ("_value", "_other")

    @property
    @override
    def _value_or_other(self) -> OtherT:
        return self._other

    @override
    def __init__(self, value: OtherT) -> None:
        self._value = value if value is undefined else undefined
        self._other = value

    @overload
    def __new__(cls, value: Undefined) -> Error[Any, Any]: ...

    @overload
    def __new__(cls, value: OtherT) -> Error[ValueT, OtherT]: ...

    @overload
    def __new__(
        cls, value: OtherT | Undefined
    ) -> Error[ValueT, OtherT] | Error[Any, Any]: ...

    @override
    def __new__(
        cls, value: OtherT | Undefined
    ) -> Error[ValueT, OtherT] | Error[Any, Any]:
        return super(Container, cls).__new__(cls)

    @override
    def map_value(self, func: Callable[[ValueT], AnotherT]) -> Result[AnotherT, OtherT]:
        return Error(self._other)

    @override
    def map_values(
        self, value: ElementT, func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Result[AnotherT, OtherT]:
        return Error(self._other)

    @override
    def map_container(
        self, value: Result[ElementT, Any], func: Callable[[ValueT, ElementT], AnotherT]
    ) -> Result[AnotherT, OtherT]:
        return Error(self._other)

    @override
    def bind_value(
        self, func: Callable[[ValueT], Result[AnotherT, OtherT]]
    ) -> Result[AnotherT, OtherT]:
        return Error(self._other)

    @override
    def bind_values(
        self,
        value: ElementT,
        func: Callable[[ValueT, ElementT], Result[AnotherT, OtherT]],
    ) -> Result[AnotherT, OtherT]:
        return Error(self._other)

    @override
    def bind_container(
        self,
        value: Result[ElementT, Any],
        func: Callable[[ValueT, ElementT], Result[AnotherT, OtherT]],
    ) -> Result[AnotherT, OtherT]:
        return Error(self._other)

    @override
    def switch(self) -> Result[OtherT, ValueT]:
        return Done(self._other)

    @override
    def default(self, value: AnotherT) -> AnotherT:
        return value

    @override
    def map_default(self, func: Callable[[], AnotherT]) -> AnotherT:
        return func()

    @override
    def unwrap(self) -> NoReturn:
        self.unwrap_error()

    @override
    def unwrap_error(self) -> NoReturn:
        if not isinstance(self._other, Exception):
            raise KontainerTypeError("error container does not hold an error")

        raise self._other

    @override
    def unwrap_error_or(self, value: AnotherT) -> AnotherT | NoReturn:
        if isinstance(self._other, Exception):
            raise self._other
        return value

    @override
    def unwrap_error_or_else(self, func: Callable[[], AnotherT]) -> AnotherT | NoReturn:
        if isinstance(self._other, Exception):
            raise self._other
        return func()

    def unwrap_error_or_with(self, func: Callable[[OtherT], AnotherT]) -> AnotherT:
        return func(self._other)

    @property
    @override
    def is_positive(self) -> Literal[False]:
        return False

    @override
    def copy(self) -> Error[ValueT, OtherT]:
        return self.error(self._other)

    @override
    def deepcopy(self) -> Error[ValueT, OtherT]:
        new = deepcopy(self._other)
        return self.error(new)

    @override
    def ensure_positive(self) -> Done[ValueT, OtherT]:
        error_msg = f"{self!r} is not done"
        raise KontainerTypeError(error_msg)

    @override
    def ensure_negative(self) -> Error[ValueT, OtherT]:
        return Error(self._other)
